use itertools::Itertools;
use serde::{Deserialize, Serialize};
use std::any::Any;
use std::collections::HashMap;
use std::collections::{BTreeMap, HashSet};
use std::fmt::{format, Error, Write};
use std::fs::File;
use std::io::BufWriter;
use std::iter::zip;
use std::mem::transmute;
use std::sync::atomic::{AtomicUsize, Ordering};

use hercules_ir::*;

use crate::*;

/** Per kernel description */

#[derive(Clone, Debug, Copy, Serialize, Deserialize)]
pub struct Switchbox<const W: usize> {
    pub output_wires: (usize, usize), // A, B
}

impl<const W: usize> Switchbox<W> {
    pub fn bits(self: &Self) -> Vec<bool> {
        let sb_bits = u32::ilog2(W.try_into().unwrap());
        let mut bits = vec![];
        let outputs = vec![self.output_wires.0, self.output_wires.1];
        for output in outputs {
            bits.append(&mut Self::sb_word_bits(output.try_into().unwrap()))
        }

        // FOR CARRIES TODO: (@XROUTH) FIX THIS.
        let mut fake_bits: Vec<bool> = (0..sb_bits).map(|a| false).collect();
        bits.append(&mut fake_bits);

        let mut fake_bits: Vec<bool> = (0..sb_bits).map(|a| false).collect();
        bits.append(&mut fake_bits);

        // bits.iter_mut().for_each(|x| *x=true);
        bits.reverse();
        bits
    }

    pub fn sb_word_bits(word: u32) -> Vec<bool> {
        let sb_bits = u32::ilog2(W.try_into().unwrap());
        (0..sb_bits)
            .rev() // Reverse to get most significant bit first
            .map(|i| (word & (1 << i)) != 0) // Check if the i-th bit is set
            .collect()
    }
}

#[derive(Clone, Debug, Copy, Serialize, Deserialize)]
pub struct FunctionalUnit {
    pub op_type: FuOp,
    pub constant: i16,
}

pub fn boolify(input: Vec<usize>) -> Vec<bool> {
    input
        .iter()
        .map(|a| if *a == 0 { false } else { true })
        .collect()
}

fn u16_to_bools(x: i16) -> Vec<bool> {
    (0..16).rev().map(|i| (x & (1 << i)) != 0).collect()
}

impl FunctionalUnit {
    pub fn bits(self: &Self) -> Vec<bool> {
        let mut bits = vec![];

        let vec = match self.op_type {
            FuOp::Add => vec![0, 0, 0, 0],
            FuOp::Sub => vec![0, 1, 1, 1],
            FuOp::Mult => vec![0, 0, 1, 0],
            FuOp::PassA => vec![1, 1, 1, 1],
            FuOp::Default => vec![1, 1, 1, 1],
        };

        bits.append(&mut boolify(vec));

        bits.push(false); // Set update acc to false.
        bits.push(false); // set acc mode to false.

        // FIXME: Do
        bits.append(&mut u16_to_bools(self.constant));

        bits.reverse();
        bits
    }
}

#[derive(Clone, Debug, Copy, Serialize, Deserialize)]
pub struct SliceDesc<const H: usize, const W: usize>
where
    [(); H]:,
    [[Switchbox<W>; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
    [[FunctionalUnit; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
{
    pub functional_units: [[FunctionalUnit; W]; H], // is 8 x 8 // Indexed [height][width]
    pub switchboxes: [[Switchbox<W>; W]; H], // 1st row of functional units doesn't have switchboxes, so this is 7 x 8
}

impl<const H: usize, const W: usize> SliceDesc<H, W>
where
    [(); H]:,
    [[Switchbox<W>; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
    [[FunctionalUnit; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
{
    pub fn pretty_print(self: Self) -> () {
        for i in 0..H - 1 {
            println!("{:?}", self.functional_units[i]);
            println!("{:?}", self.switchboxes[i]);
        }
        println!("{:?}", self.functional_units[H - 1]);
    }
}

#[derive(Clone, Debug, Copy, Serialize, Deserialize)]
pub struct ChipDesc<const H: usize, const W: usize, const K: usize>
where
    [(); H]:,
    [[Switchbox<W>; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
    [[FunctionalUnit; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
    [SliceDesc<H, W>; K]: Serialize + for<'d2> Deserialize<'d2>,
{
    pub slices: [SliceDesc<H, W>; K],
}

// Operations that individual functional units support. Unary ops operate on the first input.
#[derive(Clone, Debug, Copy, PartialEq, Serialize, Deserialize)]
pub enum FuOp {
    Add,
    Sub,
    Mult,
    PassA,
    Default, // ...
}

impl<const H: usize, const W: usize, const K: usize> ChipDesc<H, W, K>
where
    [(); H]:,
    [[Switchbox<W>; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
    [[FunctionalUnit; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
    [SliceDesc<H, W>; K]: Serialize + for<'d2> Deserialize<'d2>,
{
    pub const fn new() -> Self {
        let functional_unit = FunctionalUnit {
            op_type: FuOp::Default,
            constant: 0,
        };
        let sb = Switchbox {
            output_wires: (0, 0),
        };
        let functional_units: [[FunctionalUnit; W]; H] = [[functional_unit; W]; H];

        let slices = [SliceDesc {
            functional_units: functional_units,
            switchboxes: [[sb; W]; H],
        }; K];

        Self { slices: slices }
    }

    pub fn reduction() -> Self {
        let mut chip = ChipDesc::<H, W, K>::new();

        let REQ_NUM_ROWS = 4;

        for row in 1..H {
            let range: Box<dyn Iterator<Item = usize>> = if row % 2 == 0 {
                Box::new((0..W).rev())
            } else {
                Box::new(0..W)
            };
            // println!("range: {:?}", range.clone());

            for col in range {
                /* Determine programming word */
                if (row < REQ_NUM_ROWS) {
                    let other_wire = if col + 2_usize.pow((row - 1).try_into().unwrap()) < W {
                        col + 2_usize.pow((row - 1).try_into().unwrap())
                    } else {
                        col
                    };
                    chip.slices[0].switchboxes[row - 1][col].output_wires = (col, other_wire);

                    chip.slices[0].functional_units[row][col].op_type = FuOp::Add;
                } else {
                    chip.slices[0].switchboxes[row - 1][col].output_wires = (col, col);
                    chip.slices[0].functional_units[row][col].op_type = FuOp::PassA;
                }
            }
        }

        // Program the first row
        for col in 0..W {
            chip.slices[0].functional_units[0][col].op_type = FuOp::Add;
        }
        chip
    }
}

/*
 * The top level function to compile a Hercules IR function into LLVM IR for
 * execution on the CPU. We generate LLVM IR textually, since there are no good
 * LLVM bindings for Rust, and we are *not* writing any C++.
 */
pub fn grape_codegen<Writer: std::fmt::Write>(
    module_name: &str,
    function: &Function,
    types: &Vec<Type>,
    constants: &Vec<Constant>,
    dynamic_constants: &Vec<DynamicConstant>,
    def_use_map: &ImmutableDefUseMap,
    typing: &Vec<TypeID>,
    control_subgraph: &Subgraph,
    bbs: &BasicBlocks,
    backing_allocation: &FunctionBackingAllocation,
    w: &mut Writer,
) -> Result<(), Error> {
    let ctx = GRAPEContext::<8, 8, 1> {
        module_name,
        function,
        types,
        constants,
        dynamic_constants,
        def_use_map,
        typing,
        control_subgraph,
        bbs,
        backing_allocation,
    };
    ctx.codegen_function(w)
}

struct GRAPEContext<'a, const H: usize, const W: usize, const K: usize> {
    module_name: &'a str,
    function: &'a Function,
    types: &'a Vec<Type>,
    constants: &'a Vec<Constant>,
    dynamic_constants: &'a Vec<DynamicConstant>,
    def_use_map: &'a ImmutableDefUseMap,
    typing: &'a Vec<TypeID>,
    control_subgraph: &'a Subgraph,
    bbs: &'a BasicBlocks,
    backing_allocation: &'a FunctionBackingAllocation,
}

pub static LIB_HEADER_STR: &str = include_str!("../../hercules_rt/src/grape_header.rs");

impl<'a, const H: usize, const W: usize, const K: usize> GRAPEContext<'a, H, W, K>
where
    [(); H]:,
    [[Switchbox<W>; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
    [[FunctionalUnit; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
    [SliceDesc<H, W>; K]: Serialize + for<'d2> Deserialize<'d2>,
{
    // Remove reads

    // Parameters or constant reads from parameters.
    // Errors if not all reads are constant, or if not all parameters are arrays or constants y'know.
    fn get_initial_chip_inputs(&self) -> Result<HashSet<NodeID>, ()> {
        let mut inputs = HashSet::new();

        let parameters = self
            .function
            .nodes
            .iter()
            .enumerate()
            .filter_map(|(node_id, node)| {
                if node.is_parameter() {
                    Some(NodeID::new(node_id))
                } else {
                    None
                }
            });

        for parameter in parameters {
            let node_d = &self.function.nodes[parameter.idx()];
            let ty = &self.types[self.typing[parameter.idx()].idx()];

            match ty {
                Type::Integer16 => {
                    inputs.insert(parameter);
                }
                Type::Product(type_ids) => todo!(),
                Type::Summation(type_ids) => todo!(),
                Type::Array(type_id, dynamic_constant_ids) => {
                    let users = self.def_use_map.get_users(parameter);
                    for user in users {
                        let Node::Read {
                            collect,
                            ref indices,
                        } = &self.function.nodes[user.idx()]
                        else {
                            return Err(());
                        };
                        if indices.len() != 1 {
                            return Err(());
                        };

                        let Index::Position(pos) = &indices[0] else {
                            return Err(());
                        };

                        if pos.len() != 1 {
                            return Err(());
                        };

                        inputs.insert(*user);
                    }
                }
                _ => {}
            };
        }
        // inputs
        Ok(inputs)
    }

    fn codegen_function<Writer: Write>(&self, w: &mut Writer) -> Result<(), Error> {
        let set: HashSet<NodeID> = HashSet::from_iter(self.bbs.0.clone());

        // assert_eq!(set.len(), 2); // Because start and return are in different BBs.

        // Handle Input Nodes

        // Schedule the function.

        let return_nodes = self
            .function
            .nodes
            .iter()
            .enumerate()
            .filter_map(|(idx, node)| {
                if node.is_return() {
                    Some(NodeID::new(idx))
                } else {
                    None
                }
            });

        assert_eq!(return_nodes.clone().collect::<Vec<_>>().len(), 1);

        let returns: Vec<NodeID> = return_nodes
            .flat_map(|return_node| {
                get_uses(&self.function.nodes[return_node.idx()])
                    .as_ref()
                    .iter()
                    .filter(|node| !self.function.nodes[node.idx()].is_control())
                    .cloned()
                    .collect::<Vec<_>>()
            })
            .collect();

        let chip_inputs = self.get_initial_chip_inputs();

        let chip_inputs = chip_inputs.unwrap();

        let mut visited = HashSet::new();
        let mut stack: Vec<NodeID> = chip_inputs.iter().cloned().collect();

        while let Some(node) = stack.pop() {
            if visited.insert(node) {
                for &neighbor in get_uses(&self.function.nodes[node.idx()]).as_ref() {
                    stack.push(neighbor);
                }
            }
        }

        let mut host_nodes = HashSet::new();

        for node in visited {
            if chip_inputs.contains(&node) {
                host_nodes.insert(node);
                continue;
            } else {
                host_nodes.insert(node);
            }
        }

        assert!(chip_inputs.len() <= 16);

        println!("chip_inputs: {:?}", chip_inputs);

        let chip = self.schedule_slice_toplevel(chip_inputs); // .expect("blargh");

        let chip = if let Err((a, b, _)) = chip {
            let mut string = String::new();

            let path = "results.txt";
            let mut output = File::create(path).unwrap();

            write_xdot(&a, &mut string);

            std::fs::write("xdot.dot", string);
            todo!()
        } else {
            chip.unwrap()
        };

        // todo!();

        let (chip, output_mapping, input_mapping) = chip;
        // chip.0.pretty_print();
        // println!("chip: {:?}", chip);

        let bitstream = Self::assemble_slice(chip);

        let serialized = serde_json::to_string_pretty(&chip).unwrap();

        // // let reduction_chip = ChipDesc::<H, W, K>::reduction();
        // // let mut string = String::new();
        // // write_xdot(&reduction_chip.slices[0], &mut string);

        // // std::fs::write("xdot2.dot", string);

        // // let bitstream = Self::assemble_slice(reduction_chip.slices[0]);
        let binary_string: String = bitstream
            .iter()
            .map(|&b| if b { '1' } else { '0' })
            .collect();
        // println!("bitstream :{:?}", binary_string);
        let mut string = String::new();

        let path = "results.txt";

        write_xdot(&chip, &mut string);

        std::fs::write("xdot.dot", string);
        std::fs::write("bitstream.txt", &binary_string);

        let debug_str = format!(
            "Bitsrtearem lengh {:?} \n Input Mapping: {:?}, \n, Chip Desc: {:?}",
            binary_string.len(),
            input_mapping,
            chip
        );

        // let binary_string: String = "010100101".to_owned();
        std::fs::write("debug.txt", debug_str);

        // todo!();

        // write!(output, "{}", string);

        // todo!();
        // Assemble to bitstream.
        // todo!();

        self.codegen_launch_rust_function(
            binary_string,
            Some(serialized),
            input_mapping,
            output_mapping,
            returns,
            self.bbs,
            host_nodes,
            w,
        )?;

        Ok(())
    }

    fn get_type(&self, id: TypeID) -> String {
        Self::convert_type(&self.types[id.idx()], &self.types)
    }

    fn convert_type(ty: &Type, types: &[Type]) -> String {
        match ty {
            Type::Boolean => "bool".to_string(),
            Type::Integer8 => "i8".to_string(),
            Type::Integer16 => "i16".to_string(),
            Type::Integer32 => "i32".to_string(),
            Type::Integer64 => "i64".to_string(),
            Type::UnsignedInteger8 => "u8".to_string(),
            Type::UnsignedInteger16 => "u16".to_string(),
            Type::UnsignedInteger32 => "u32".to_string(),
            Type::UnsignedInteger64 => "u64".to_string(),
            Type::Float32 => "f32".to_string(),
            Type::Float64 => "f64".to_string(),
            Type::Product(_) | Type::Summation(_) | Type::Array(_, _) => {
                "::hercules_rt::__RawPtrSendSync".to_string()
            }
            Type::MultiReturn(ts) => {
                format!(
                    "({})",
                    ts.iter()
                        .map(|t| Self::convert_type(&types[t.idx()], types))
                        .collect::<Vec<_>>()
                        .join(", ")
                )
            }
            _ => panic!(),
        }
    }

    // This can be rust or cpp or llvm, who cares,
    // for now make it rust because that is interesting i guess
    fn codegen_launch_rust_function<Writer: Write>(
        &self,
        bitstream: String,
        json_chipdesc: Option<String>,
        input_mapping: Vec<NodeID>,
        output_mapping: HashMap<NodeID, usize>,
        returns: Vec<NodeID>,
        bbs: &BasicBlocks,
        host_nodes: HashSet<NodeID>,
        w: &mut Writer,
    ) -> Result<(), Error> {
        // Write bitstream;
        let bistream_name = format!("BITSTREAM_{}_{}", self.module_name, self.function.name);

        // write!(w, "#[cfg(feature = \"grape\")] \n")?;
        if !json_chipdesc.is_some() {
            write!(
                w,
                "static {}: &[u8] = {}\n",
                bistream_name,
                Self::embed_bitstream(&bitstream)
            )?;
        } else {
            write!(
                w,
                "static {}: &str = r#\"{}\"#;\n",
                bistream_name,
                json_chipdesc.clone().unwrap(),
            )?;
        }

        let is_multi_return = self.function.return_types.len() != 1;

        if is_multi_return {
            write!(
                w,
                "#[repr(C)] struct ReturnStruct{} {{ {} }} \n",
                self.function.name,
                self.function
                    .return_types
                    .iter()
                    .enumerate()
                    .map(|(idx, t)| format!("f{}: {}", idx, self.get_type(*t)))
                    .collect::<Vec<_>>()
                    .join(", "),
            )?;
        }
        // write!(w, "#[cfg(feature = \"grape\")] \n")?;
        write!(w, "#[no_mangle]\n")?;
        write!(
            w,
            "pub extern \"C\" fn {}_{}(",
            self.module_name, self.function.name
        )?;

        // Parameters
        // write!(w, "backing_ptr: *mut u8")?;
        let mut first_param = true;
        // if self.backing_allocations[&func_id].contains_key(&self.devices[func_id.idx()]) {
        //     first_param = false;
        //     write!(w, "backing: ::hercules_rt::__RawPtrSendSync")?;
        // }
        for idx in 0..self.function.num_dynamic_constants {
            if first_param {
                first_param = false;
            } else {
                write!(w, ", ")?;
            }
            write!(w, "dc{}: u64", idx)?;
        }
        for (idx, ty) in self.function.param_types.iter().enumerate() {
            if first_param {
                first_param = false;
            } else {
                write!(w, ", ")?;
            }
            write!(w, "p{}: {}", idx, self.get_type(*ty))?;
        }
        // return type

        if !is_multi_return {
            write!(w, ") -> {}", self.get_type(self.function.return_types[0]));
        } else {
            if first_param {
                first_param = false;
            } else {
                write!(w, ", ")?;
            }
            write!(w, "ret_ptr: *mut ReturnStruct{})", self.function.name);
        }

        // function body
        write!(w, "\n{{\n")?;

        write!(w, "unsafe {{\n")?;

        // TODO: first time only programming

        let bitstream_length = 2080;
        let bitstream_length = json_chipdesc.unwrap().len();

        write!(
            w,
            "program_bitstream({}.as_ptr(), {bitstream_length});\n ",
            bistream_name
        )?;

        write!(w, "let mut input: [i16; 16] = [0; 16]; \n ")?;
        write!(w, "let mut output: [i16; 8] = [0; 8]; \n ")?;

        // Iterate parameter mapping
        println!("input_mapping: {:?}", input_mapping);
        // panic!();

        for block in self.bbs.1.iter() {
            for id in block {
                if host_nodes.contains(id) {
                    write!(w, "let ");
                    self.codegen_data_node(*id, w)?;
                    write!(w, "\n");
                }
            }
        }

        for (idx, node) in input_mapping.iter().enumerate() {
            if node.idx() != 0 {
                let rhs = self.get_value(*node);

                write!(w, "input[{idx}] = {};\n", rhs);
            }

            // let rhs = match &self.function.nodes[node.idx()] {
            //     Node::Parameter { index } => {
            //         format!("p{}", index)
            //     }
            //     Node::Read { collect, indices } => {
            //         // Figure out what parameter the collection is, if any. IT HAS TO BE ONE
            //         //Do the read in this function.

            //         // FIXME: just asssume its p0 for now
            //         let Index::Position(i) = &indices[0] else {
            //             panic!();
            //         };

            //         let Node::Constant { id } = self.function.nodes[i[0].idx()] else {
            //             panic!();
            //         };
            //         let Constant::UnsignedInteger64(v) = self.constants[id.idx()] else {
            //             panic!()
            //         };

            //         format!("*((p0.0 as *const i16).add({}))", v)
            //     }
            //     Node::Start => "".to_owned(),
            //     _ => panic!(),
            // };
            // skip start node
        }
        write!(w, "grape_sim::send_data(input.as_ptr());\n")?;
        write!(w, "await_valid();\n")?;
        write!(w, "read_data(output.as_mut_ptr());\n")?;

        // Iterate return value mapping.

        if is_multi_return {
            for (idx, node) in returns.iter().enumerate() {
                write!(
                    w,
                    "(*ret_ptr).f{} = output[{}];\n",
                    idx, output_mapping[node]
                );
            }
            write!(w, "return;\n")?;
        } else {
            write!(w, "return output[{}];\n", output_mapping[&returns[0]])?;
        }

        // close unsafe
        write!(w, "}}\n")?;
        write!(w, "}}")?;

        Ok(())
    }

    fn embed_bitstream(bitstream: &str) -> String {
        let mut output = String::from("&[\n");
        let mut current_byte = 0u8;
        let mut bits_collected = 0;

        for c in bitstream.chars() {
            current_byte <<= 1;
            match c {
                '0' => {}
                '1' => current_byte |= 1,
                _ => panic!("Invalid character in bitstream: {}", c),
            }
            bits_collected += 1;

            if bits_collected == 8 {
                output.push_str(&format!("    0b{:08b},\n", current_byte));
                current_byte = 0;
                bits_collected = 0;
            }
        }

        if bits_collected > 0 {
            current_byte <<= 8 - bits_collected;
            output.push_str(&format!("    0b{:08b},\n", current_byte));
        }

        output.push_str("];\n");
        output
    }

    fn get_value(&self, id: NodeID) -> String {
        format!("node_{}", id.idx())
    }

    fn codegen_data_node<Writer: Write>(&self, id: NodeID, w: &mut Writer) -> Result<(), Error> {
        let func = &self.function;
        match func.nodes[id.idx()] {
            Node::Parameter { index } => write!(w, "{} = p{};", self.get_value(id), index)?,
            Node::Constant { id: cons_id } => {
                write!(w, "{} = ", self.get_value(id))?;
                match self.constants[cons_id.idx()] {
                    Constant::Boolean(val) => write!(w, "{}", val)?,
                    Constant::Integer8(val) => write!(w, "{}i8", val)?,
                    Constant::Integer16(val) => write!(w, "{}i16", val)?,
                    Constant::Integer32(val) => write!(w, "{}i32", val)?,
                    Constant::Integer64(val) => write!(w, "{}i64", val)?,
                    Constant::UnsignedInteger8(val) => write!(w, "{}u8", val)?,
                    Constant::UnsignedInteger16(val) => write!(w, "{}u16", val)?,
                    Constant::UnsignedInteger32(val) => write!(w, "{}u32", val)?,
                    Constant::UnsignedInteger64(val) => write!(w, "{}u64", val)?,
                    Constant::Float32(val) => {
                        if val == f32::INFINITY {
                            write!(w, "f32::INFINITY")?
                        } else if val == f32::NEG_INFINITY {
                            write!(w, "f32::NEG_INFINITY")?
                        } else {
                            write!(w, "{}f32", val)?
                        }
                    }
                    Constant::Float64(val) => {
                        if val == f64::INFINITY {
                            write!(w, "f64::INFINITY")?
                        } else if val == f64::NEG_INFINITY {
                            write!(w, "f64::NEG_INFINITY")?
                        } else {
                            write!(w, "{}f64", val)?
                        }
                    }
                    Constant::Product(ty, _)
                    | Constant::Summation(ty, _, _)
                    | Constant::Array(ty) => {
                        todo!()
                    }
                }
                write!(w, ";")?;
            }
            Node::DynamicConstant { id: dc_id } => {
                todo!()
            }
            Node::Unary { op, input } => {
                match op {
                    UnaryOperator::Not => {
                        write!(w, "{} = !{};", self.get_value(id), self.get_value(input))?
                    }
                    UnaryOperator::Neg => {
                        write!(w, "{} = -{};", self.get_value(id), self.get_value(input))?
                    }
                    UnaryOperator::Cast(ty) => write!(
                        w,
                        "{} = {} as {};",
                        self.get_value(id),
                        self.get_value(input),
                        self.get_type(ty)
                    )?,
                };
            }
            Node::Binary { op, left, right } => {
                let op = match op {
                    BinaryOperator::Add => "+",
                    BinaryOperator::Sub => "-",
                    BinaryOperator::Mul => "*",
                    BinaryOperator::Div => "/",
                    BinaryOperator::Rem => "%",
                    BinaryOperator::LT => "<",
                    BinaryOperator::LTE => "<=",
                    BinaryOperator::GT => ">",
                    BinaryOperator::GTE => ">=",
                    BinaryOperator::EQ => "==",
                    BinaryOperator::NE => "!=",
                    BinaryOperator::Or => "|",
                    BinaryOperator::And => "&",
                    BinaryOperator::Xor => "^",
                    BinaryOperator::LSh => "<<",
                    BinaryOperator::RSh => ">>",
                };

                write!(
                    w,
                    "{} = {} {} {};",
                    self.get_value(id),
                    self.get_value(left),
                    op,
                    self.get_value(right)
                )?;
            }
            Node::Ternary {
                op,
                first,
                second,
                third,
            } => {
                match op {
                    TernaryOperator::Select => write!(
                        w,
                        "{} = if {} {{{}}} else {{{}}};",
                        self.get_value(id),
                        self.get_value(first),
                        self.get_value(second),
                        self.get_value(third),
                    )?,
                };
            }
            Node::Read {
                collect,
                ref indices,
            } => {
                let collect_ty = self.typing[collect.idx()];
                let self_ty = self.typing[id.idx()];
                let offset = self.codegen_index_math(collect_ty, indices)?;
                if self.types[self_ty.idx()].is_primitive() {
                    write!(
                        w,
                        "{} = ({}.byte_add({} as usize).0 as *mut {}).read();",
                        self.get_value(id),
                        self.get_value(collect),
                        offset,
                        self.get_type(self_ty)
                    )?;
                } else {
                    write!(
                        w,
                        "{} = {}.byte_add({} as usize);",
                        self.get_value(id),
                        self.get_value(collect),
                        offset,
                    )?;
                }
            }
            Node::Write {
                collect,
                data,
                ref indices,
            } => {
                todo!()
            }
            _ => panic!(
                "PANIC: Can't lower {:?} in {}.",
                func.nodes[id.idx()],
                func.name
            ),
        }
        Ok(())
    }

    fn codegen_type_size(&self, ty: TypeID) -> String {
        match self.types[ty.idx()] {
            Type::Control | Type::MultiReturn(_) => panic!(),
            Type::Boolean | Type::Integer8 | Type::UnsignedInteger8 | Type::Float8 => {
                "1".to_string()
            }
            Type::Integer16 | Type::UnsignedInteger16 | Type::BFloat16 => "2".to_string(),
            Type::Integer32 | Type::UnsignedInteger32 | Type::Float32 => "4".to_string(),
            Type::Integer64 | Type::UnsignedInteger64 | Type::Float64 => "8".to_string(),
            Type::Product(ref fields) => {
                let fields_align = fields
                    .into_iter()
                    .map(|id| get_type_alignment(&self.types, *id));
                let fields: Vec<String> = fields
                    .into_iter()
                    .map(|id| self.codegen_type_size(*id))
                    .collect();

                // Emit LLVM IR to round up to the alignment of the next field,
                // and then add the size of that field. At the end, round up to
                // the alignment of the whole struct.
                let mut acc_size = "0".to_string();
                for (field_align, field) in zip(fields_align, fields) {
                    acc_size = format!(
                        "(({} + {}) & !{})",
                        acc_size,
                        field_align - 1,
                        field_align - 1
                    );
                    acc_size = format!("({} + {})", acc_size, field);
                }
                let total_align = get_type_alignment(&self.types, ty);
                format!(
                    "(({} + {}) & !{})",
                    acc_size,
                    total_align - 1,
                    total_align - 1
                )
            }
            Type::Summation(ref variants) => {
                let variants = variants.into_iter().map(|id| self.codegen_type_size(*id));

                // The size of a summation is the size of the largest field,
                // plus 1 byte and alignment for the discriminant.
                let mut acc_size = "0".to_string();
                for variant in variants {
                    acc_size = format!("::core::cmp::max({}, {})", acc_size, variant);
                }

                // No alignment is necessary before the 1 byte discriminant.
                let total_align = get_type_alignment(&self.types, ty);
                format!(
                    "(({} + 1 + {}) & !{})",
                    acc_size,
                    total_align - 1,
                    total_align - 1
                )
            }
            Type::Array(elem, ref bounds) => {
                // The size of an array is the size of the element multipled by
                // the dynamic constant bounds.
                let mut acc_size = self.codegen_type_size(elem);
                let elem_align = get_type_alignment(&self.types, elem);
                acc_size = format!(
                    "(({} + {}) & !{})",
                    acc_size,
                    elem_align - 1,
                    elem_align - 1
                );
                for dc in bounds {
                    acc_size = format!("{} * ", acc_size);
                    self.codegen_dynamic_constant(*dc, &mut acc_size).unwrap();
                }
                format!("({})", acc_size)
            }
        }
    }

    /*
     * Lower dynamic constant in Hercules IR into a Rust expression.
     */
    fn codegen_dynamic_constant<Writer: Write>(
        &self,
        id: DynamicConstantID,
        w: &mut Writer,
    ) -> Result<(), Error> {
        match &self.dynamic_constants[id.idx()] {
            DynamicConstant::Constant(val) => write!(w, "{}", val)?,
            DynamicConstant::Parameter(idx) => write!(w, "dc_p{}", idx)?,
            DynamicConstant::Add(xs) => {
                write!(w, "(")?;
                let mut xs = xs.iter();
                self.codegen_dynamic_constant(*xs.next().unwrap(), w)?;
                for x in xs {
                    write!(w, "+")?;
                    self.codegen_dynamic_constant(*x, w)?;
                }
                write!(w, ")")?;
            }
            DynamicConstant::Sub(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(*left, w)?;
                write!(w, "-")?;
                self.codegen_dynamic_constant(*right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Mul(xs) => {
                write!(w, "(")?;
                let mut xs = xs.iter();
                self.codegen_dynamic_constant(*xs.next().unwrap(), w)?;
                for x in xs {
                    write!(w, "*")?;
                    self.codegen_dynamic_constant(*x, w)?;
                }
                write!(w, ")")?;
            }
            DynamicConstant::Div(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(*left, w)?;
                write!(w, "/")?;
                self.codegen_dynamic_constant(*right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Rem(left, right) => {
                write!(w, "(")?;
                self.codegen_dynamic_constant(*left, w)?;
                write!(w, "%")?;
                self.codegen_dynamic_constant(*right, w)?;
                write!(w, ")")?;
            }
            DynamicConstant::Min(xs) => {
                let mut xs = xs.iter().peekable();

                // Track the number of parentheses we open that need to be closed later
                let mut opens = 0;
                while let Some(x) = xs.next() {
                    if xs.peek().is_none() {
                        // For the last element, we just print it
                        self.codegen_dynamic_constant(*x, w)?;
                    } else {
                        // Otherwise, we create a new call to min and print the element as the
                        // first argument
                        write!(w, "::core::cmp::min(")?;
                        self.codegen_dynamic_constant(*x, w)?;
                        write!(w, ",")?;
                        opens += 1;
                    }
                }
                for _ in 0..opens {
                    write!(w, ")")?;
                }
            }
            DynamicConstant::Max(xs) => {
                let mut xs = xs.iter().peekable();

                let mut opens = 0;
                while let Some(x) = xs.next() {
                    if xs.peek().is_none() {
                        self.codegen_dynamic_constant(*x, w)?;
                    } else {
                        write!(w, "::core::cmp::max(")?;
                        self.codegen_dynamic_constant(*x, w)?;
                        write!(w, ",")?;
                        opens += 1;
                    }
                }
                for _ in 0..opens {
                    write!(w, ")")?;
                }
            }
        }
        Ok(())
    }

    /*
     * Emit logic to index into an collection.
     */
    fn codegen_index_math(
        &self,
        mut collect_ty: TypeID,
        indices: &[Index],
    ) -> Result<String, Error> {
        let mut acc_offset = "0".to_string();
        for index in indices {
            match index {
                Index::Field(idx) => {
                    let Type::Product(ref fields) = self.types[collect_ty.idx()] else {
                        panic!()
                    };

                    // Get the offset of the field at index `idx` by calculating
                    // the product's size up to field `idx`, then offseting the
                    // base pointer by that amount.
                    for field in &fields[..*idx] {
                        let field_align = get_type_alignment(&self.types, *field);
                        let field = self.codegen_type_size(*field);
                        acc_offset = format!(
                            "((({} + {}) & !{}) + {})",
                            acc_offset,
                            field_align - 1,
                            field_align - 1,
                            field
                        );
                    }
                    let last_align = get_type_alignment(&self.types, fields[*idx]);
                    acc_offset = format!(
                        "(({} + {}) & !{})",
                        acc_offset,
                        last_align - 1,
                        last_align - 1
                    );
                    collect_ty = fields[*idx];
                }
                Index::Variant(idx) => {
                    // The tag of a summation is at the end of the summation, so
                    // the variant pointer is just the base pointer. Do nothing.
                    let Type::Summation(ref variants) = self.types[collect_ty.idx()] else {
                        panic!()
                    };
                    collect_ty = variants[*idx];
                }
                Index::Position(ref pos) => {
                    let Type::Array(elem, ref dims) = self.types[collect_ty.idx()] else {
                        panic!()
                    };

                    // The offset of the position into an array is:
                    //
                    //     ((0 * s1 + p1) * s2 + p2) * s3 + p3 ...
                    let elem_size = self.codegen_type_size(elem);
                    let elem_align = get_type_alignment(&self.types, elem);
                    let aligned_elem_size = format!(
                        "(({} + {}) & !{})",
                        elem_size,
                        elem_align - 1,
                        elem_align - 1
                    );
                    for (p, s) in zip(pos, dims) {
                        let p = self.get_value(*p);
                        acc_offset = format!("{} * ", acc_offset);
                        self.codegen_dynamic_constant(*s, &mut acc_offset)?;
                        acc_offset = format!("({} + {})", acc_offset, p);
                    }

                    // Convert offset in # elements -> # bytes.
                    acc_offset = format!("({} * {})", acc_offset, aligned_elem_size);
                    collect_ty = elem;
                }
            }
        }
        Ok(acc_offset)
    }

    fn assemble_slice(slice: SliceDesc<H, W>) -> Vec<bool> {
        // First row of FUs
        let mut bits = vec![];

        for i in 0..W {
            let mut bits_for_fu = slice.functional_units[0][i].bits();
            bits.append(&mut bits_for_fu);
        }

        for i in 1..H {
            println!("i: {:?}", i);
            let range: Box<dyn Iterator<Item = usize>> = if i % 2 == 1 {
                Box::new((0..W).rev())
            } else {
                Box::new(0..W)
            };
            // println!("range: {:?}", range.clone());

            for j in range {
                // println!("j: {:?}", j);

                let sb = slice.switchboxes[i - 1][j];
                let fu = slice.functional_units[i][j];

                // Switchbox,
                let mut bits_for_sb = sb.bits();
                // println!("bits_for_sb: {:?}", bits_for_sb);

                // println!("bits_for_sb: {:?}", bits_for_sb.len());

                // FU
                let mut bits_for_fu = fu.bits();
                // println!("bits_for_fu: {:?}", bits_for_fu);
                // println!("bits_for_fu: {:?}", bits_for_fu.len());

                bits.append(&mut bits_for_sb);
                bits.append(&mut bits_for_fu);
            }
        }
        // panic!();
        bits.reverse();
        bits
    }

    fn get_free_col(live_outs: &HashMap<NodeID, usize>) -> usize {
        (0..W)
            .find(|col| {
                live_outs
                    .values()
                    .cloned()
                    .find(|value| value == col)
                    .is_none()
            })
            .unwrap()
    }

    fn schedule_slice_toplevel(
        &self,
        input_nodes: HashSet<NodeID>,
    ) -> Result<
        (SliceDesc<H, W>, HashMap<NodeID, usize>, Vec<NodeID>),
        (SliceDesc<H, W>, HashMap<NodeID, usize>, Vec<NodeID>),
    > {
        let mut functional_units = [[FunctionalUnit {
            op_type: FuOp::Default,
            constant: 0,
        }; W]; H];
        let mut switchboxes = [[Switchbox {
            output_wires: (100, 100),
        }; W]; H];

        let config = SliceDesc {
            functional_units,
            switchboxes,
        };

        let input_mapping = vec![NodeID::new(0); W * 2];

        let mut param_map = HashMap::new();
        for (i, item) in input_nodes.iter().enumerate() {
            param_map.insert(*item, i);
        }
        return self.schedule_row_recursive(0, param_map, input_mapping, config, HashSet::new());
    }

    fn schedule_row_recursive(
        &self,
        row: usize,
        prev_mapping: HashMap<NodeID, usize>, // A node mapping for the previous row
        mut input_mapping: Vec<NodeID>,
        mut config: SliceDesc<H, W>,
        mut computed_nodes: HashSet<NodeID>,
    ) -> Result<
        (SliceDesc<H, W>, HashMap<NodeID, usize>, Vec<NodeID>),
        (SliceDesc<H, W>, HashMap<NodeID, usize>, Vec<NodeID>),
    > {
        println!("entry @ row {:?}", row);
        println!("prev_mapping: {:?}", prev_mapping);

        if row == H {
            // all uses of any return ndoes are in the prev mapping.
            println!("final row");
            let returns = self
                .function
                .nodes
                .iter()
                .enumerate()
                .filter_map(|(idx, node)| {
                    if node.is_return() {
                        Some(NodeID::new(idx))
                    } else {
                        None
                    }
                });

            let uses: Vec<NodeID> = returns
                .flat_map(|return_node| {
                    get_uses(&self.function.nodes[return_node.idx()])
                        .as_ref()
                        .iter()
                        .filter(|node| !self.function.nodes[node.idx()].is_control())
                        .cloned()
                        .collect::<Vec<_>>()
                })
                .collect();

            println!("return_uses: {:?}", uses.clone());

            if uses.iter().any(|node| !prev_mapping.contains_key(node)) {
                println!("Bad!");
                return Err((config, prev_mapping, input_mapping));
            } else {
                return Ok((config, prev_mapping, input_mapping));
            }
        }

        let constants: Vec<NodeID> = self
            .function
            .nodes
            .iter()
            .enumerate()
            .filter_map(|(id, data)| {
                if data.is_constant() {
                    let Node::Constant { id: c_id } = data else {
                        return None;
                    };
                    let constant = &self.constants[c_id.idx()];
                    let Constant::Integer16(value) = constant else {
                        return None;
                    };
                    Some(NodeID::new(id))
                } else {
                    None
                }
            })
            .collect();
        // TODO: PRUNE constants that can't possibly be used

        // PRUNE nodes in prev that don't have any users
        let live_nodes: Vec<NodeID> = prev_mapping
            .iter()
            .filter_map(|(id, _)| {
                if self.def_use_map.get_users(*id).iter().count() > 0 {
                    Some(*id)
                } else {
                    None
                }
            })
            .collect();

        //
        let users = prev_mapping
            .iter()
            .flat_map(|(id, _)| self.def_use_map.get_users(*id).iter().cloned());

        // Correctness:
        // PRUNE nodes that don't have all their uses computed.
        let users: Vec<NodeID> = users
            .filter(|id| {
                get_uses(&self.function.nodes[id.idx()])
                    .as_ref()
                    .iter()
                    .filter(|node| !self.function.nodes[node.idx()].is_control()) // We only care abt data nodes.
                    .all(|u| prev_mapping.contains_key(u))
            })
            .collect();

        // Heurestic (Speed): PRIORITIZE nodes that don't have all their users scheduled.
        // we should first choose the set of computations of we want to do, and then the set of values
        // that we want to keep live
        // Q: Why not FORCE?, A: we can rematerialize, and sometimes that is correct in the optimal mapping
        // We can also do this by sorting live_nodes by number of users (that aren't scheduled.)
        // Do we have to keep track of everything we scheduled then? Maybe.

        // Put users first, to prioritize new computation.
        let mut choices: Vec<NodeID> = users
            .into_iter()
            .chain(constants.into_iter())
            .chain(live_nodes.into_iter())
            .filter(|choice| !self.function.nodes[choice.idx()].is_return())
            .collect();

        fn dedup<T: Eq + std::hash::Hash + Clone>(vec: &mut Vec<T>) {
            let mut seen = HashSet::new();
            vec.retain(|item| seen.insert(item.clone()));
        }

        dedup(&mut choices);
        // Remove duplicates without changing order.

        // Move killed nodes to back.
        let mut i = 0;

        println!("computed_nodes: {:?}", computed_nodes);

        choices = choices
            .iter()
            .filter(|i| {
                !self
                    .def_use_map
                    .get_users(**i)
                    .iter()
                    .all(|u| computed_nodes.contains(u))
            })
            .cloned()
            .collect();

        println!("num choices: {:?}", choices.clone().len());
        println!("choices: {:?}", choices.clone());

        // Order by node id, and start iterating permutations (NODES choose 8) for the next row.
        for combo in choices.iter().combinations(W.min(choices.len())) {
            println!("combo: {:?}", combo);
            let mut next_mapping = HashMap::new();
            let mut next_computed_nodes = computed_nodes.clone();

            // Heurestic (Correctness): Just assume the first 8 are good choices.
            for (col, node) in combo.iter().take(W).enumerate() {
                // If the node was in the previous mapping and we chose to live, schedule it as a pass through
                if prev_mapping.contains_key(node) {
                    next_mapping.insert(**node, col);

                    // Collect inputs
                    if row != 0 {
                        config.switchboxes[row - 1][col].output_wires = (prev_mapping[node], 0);
                    } else {
                        // If we are  in the initial row, put this data on the input wire
                        input_mapping[col * 2] = **node;
                    }
                    config.functional_units[row][col].op_type = FuOp::PassA;
                } else if self.function.nodes[node.idx()].is_constant() {
                    next_mapping.insert(**node, col);
                    let Node::Constant { id } = self.function.nodes[node.idx()] else {
                        panic!()
                    };
                    let constant = &self.constants[id.idx()];
                    let Constant::Integer16(value) = constant else {
                        panic!()
                    };
                    config.functional_units[row][col].constant = *value;
                } else {
                    // compute it
                    next_mapping.insert(**node, col);
                    next_computed_nodes.insert(**node);

                    // Collect inputs
                    let inputs: Vec<usize> = get_uses(&self.function.nodes[node.idx()])
                        .as_ref()
                        .iter()
                        .map(|node| prev_mapping[node])
                        .collect();

                    if row != 0 {
                        config.switchboxes[row - 1][col].output_wires = (
                            *(inputs.get(0).unwrap_or(&0)),
                            *(inputs.get(1).unwrap_or(&0)),
                        );
                    } else {
                        let node_inputs: Vec<NodeID> = get_uses(&self.function.nodes[node.idx()])
                            .as_ref()
                            .iter()
                            .cloned()
                            .collect();
                        // If we are  in the initial row, put the desired inputs on the input wires.
                        input_mapping[col * 2] = *node_inputs.get(0).unwrap_or(&NodeID::new(0));
                        input_mapping[col * 2 + 1] = *node_inputs.get(1).unwrap_or(&NodeID::new(0));
                    }

                    let op_type = match self.function.nodes[node.idx()] {
                        Node::Binary { left, right, op } => match op {
                            BinaryOperator::Add => FuOp::Add,
                            BinaryOperator::Sub => FuOp::Sub,
                            BinaryOperator::Mul => FuOp::Mult,
                            BinaryOperator::Div => todo!(),
                            BinaryOperator::Rem => todo!(),
                            BinaryOperator::LT => todo!(),
                            BinaryOperator::LTE => todo!(),
                            BinaryOperator::GT => todo!(),
                            BinaryOperator::GTE => todo!(),
                            BinaryOperator::EQ => todo!(),
                            BinaryOperator::NE => todo!(),
                            BinaryOperator::Or => todo!(),
                            BinaryOperator::And => todo!(),
                            BinaryOperator::Xor => todo!(),
                            BinaryOperator::LSh => todo!(),
                            BinaryOperator::RSh => todo!(),
                        },
                        _ => todo!(),
                    };
                    config.functional_units[row][col].op_type = op_type; // FIXME:
                }
            }

            let schedule_attempt = self.schedule_row_recursive(
                row + 1,
                next_mapping,
                input_mapping.clone(),
                config,
                next_computed_nodes.clone(),
            );

            if schedule_attempt.is_ok() {
                return schedule_attempt;
            }
        }

        return Err((config, prev_mapping, input_mapping));
    }

    // fn schedule_slice(
    //     &self,
    //     mut live_outs: HashMap<NodeID, usize>,
    // ) -> Result<(SliceDesc<H, W>, HashMap<NodeID, usize>), Error> {
    //     let mut next_live_outs: HashMap<NodeID, usize> = HashMap::new(); // Node to the col they are live in.
    //     let mut live_not_computed: HashSet<NodeID> = HashSet::new();

    //     let mut functional_units = [[FunctionalUnit {
    //         op_type: FuOp::Default,
    //         constant: 0,
    //     }; W]; H];
    //     let mut switchboxes = [[Switchbox {
    //         output_wires: (100, 100),
    //     }; W]; H - 1];

    //     // let mut next_live_outs: [NodeID; W] = [NodeID::new(0); W];
    //     // Assign to last row
    //     for row in (0..H).rev() {
    //         println!("row: {:?}", row);
    //         next_live_outs.clear();

    //         let mut computed_nodes: HashSet<NodeID> = HashSet::new();
    //         // Schedule not computed nodes.
    //         for not_computed_node in &live_not_computed {
    //             println!("in live not computed node {:?}", not_computed_node);
    //             let mut users = self
    //                 .def_use_map
    //                 .get_users(*not_computed_node)
    //                 .iter()
    //                 .filter(|node| !self.function.nodes[node.idx()].is_control());

    //             let col;
    //             if !users.all(|user| live_outs.contains_key(user)) {
    //                 // Can't schedule this yet, schedule a passthrough instead
    //                 col = Self::get_free_col(&next_live_outs);
    //                 next_live_outs.insert(*not_computed_node, col);
    //                 // Schedule Pass Through
    //                 functional_units[row][col] = FunctionalUnit {
    //                     op_type: FuOp::PassA,
    //                 };
    //                 println!(
    //                     "schedule node {:?} pass through  @ {:?}",
    //                     not_computed_node, col
    //                 );
    //             } else {
    //                 // Schedule Computation
    //                 col = Self::get_free_col(&next_live_outs);
    //                 next_live_outs.insert(*not_computed_node, col);
    //                 computed_nodes.insert(not_computed_node.clone());

    //                 functional_units[row][col] = FunctionalUnit { op_type: FuOp::Add };
    //                 println!(
    //                     "schedule node {:?} computation  @ {:?}",
    //                     not_computed_node, col
    //                 );

    //                 // live_not_computed.remove(&not_computed_node);
    //             }

    //             let live_out_col = live_outs[&not_computed_node];

    //             // Propogate to itself
    //             switchboxes[row][live_out_col] = Switchbox {
    //                 output_wires: (col, col),
    //             };

    //             continue;
    //         }

    //         for (live_out, live_out_col) in &live_outs {
    //             // If this node is dead, don't schedule
    //             if live_not_computed.contains(live_out) {
    //                 continue;
    //             }

    //             // If live outs, schedule as pass throughs to top.
    //             let node = &self.function.nodes[live_out.idx()];
    //             println!(
    //                 "live_out_node: {:?} id: {:?}, col: {:?}",
    //                 node, live_out, live_out_col
    //             );

    //             let mut uses: Vec<NodeID> = get_uses(&self.function.nodes[live_out.idx()])
    //                 .as_ref()
    //                 .iter()
    //                 .filter(|n| !self.function.nodes[n.idx()].is_control())
    //                 .cloned()
    //                 .collect();

    //             if (node.is_constant() || node.is_dynamic_constant() || node.is_parameter()) {
    //                 uses.push(*live_out);
    //             }

    //             let mut idx_vec = [3, 3];
    //             let mut ctr = 0;
    //             println!("uses: {:?}", uses);

    //             // Try to schedule things live_outs uses.
    //             for u in uses {
    //                 let users = self.def_use_map.get_users(u);
    //                 let u_node_data = &self.function.nodes[u.idx()];

    //                 if next_live_outs.contains_key(&u) {
    //                     // reuse
    //                     if row != (H - 1) {
    //                         idx_vec[ctr] = next_live_outs[&u];
    //                     }
    //                     println!("reuse value {:?}", u);
    //                 } else if live_outs.contains_key(&u) {
    //                     // Keep it live.
    //                     let col = Self::get_free_col(&next_live_outs);

    //                     next_live_outs.insert(u, col);
    //                     functional_units[row][col] = FunctionalUnit {
    //                         op_type: FuOp::PassA,
    //                     };
    //                     println!(
    //                         "schedule already live value {:?} pass through  @ {:?}",
    //                         u, col
    //                     );
    //                     if row != (H - 1) {
    //                         idx_vec[ctr] = col;
    //                     }
    //                 } else if (u_node_data.is_constant()
    //                     || u_node_data.is_dynamic_constant()
    //                     || u_node_data.is_parameter())
    //                 {
    //                     let col = Self::get_free_col(&next_live_outs);
    //                     next_live_outs.insert(u, col);
    //                     // Schedule Pass Through
    //                     functional_units[row][col] = FunctionalUnit {
    //                         op_type: FuOp::PassA,
    //                     };
    //                     println!("schedule parameter {:?} pass through  @ {:?}", u, col);
    //                     if row != (H - 1) {
    //                         idx_vec[ctr] = col;
    //                     }
    //                 } else if !users.iter().all(|user| live_outs.contains_key(user)) {
    //                     // Can't schedule this yet, schedule a passthrough instead
    //                     let col = Self::get_free_col(&next_live_outs);
    //                     next_live_outs.insert(u, col);
    //                     live_not_computed.insert(u);
    //                     // Schedule Pass Through
    //                     functional_units[row][col] = FunctionalUnit {
    //                         op_type: FuOp::PassA,
    //                     };
    //                     println!("schedule node {:?} pass through  @ {:?}", u, col);
    //                     if row != (H - 1) {
    //                         idx_vec[ctr] = col;
    //                     }
    //                 } else {
    //                     // Schedule Computation
    //                     let col = Self::get_free_col(&next_live_outs);
    //                     next_live_outs.insert(u, col);

    //                     functional_units[row][col] = FunctionalUnit { op_type: FuOp::Add };
    //                     println!("schedule node {:?} computation  @ {:?}", u, col);
    //                     if row != (H - 1) {
    //                         idx_vec[ctr] = col;
    //                     }

    //                     live_not_computed.remove(&u);
    //                 }
    //                 ctr += 1;
    //             }

    //             // Set switcboxes for newly scheduled inputs.

    //             // maybe row + 1.
    //             if row != (H - 1) {
    //                 let a = idx_vec[0];
    //                 let b = idx_vec[1];
    //                 // According to binop type

    //                 println!(
    //                     "setting switchbox @ r {:?}, c {:?} to {:?}",
    //                     row,
    //                     *live_out_col,
    //                     (a, b)
    //                 );
    //                 switchboxes[row][*live_out_col] = Switchbox {
    //                     output_wires: (a, b),
    //                 };
    //             }
    //         }

    //         live_outs = next_live_outs.clone();

    //         for node in computed_nodes {
    //             live_not_computed.remove(&node);
    //         }
    //     }

    //     Ok((
    //         SliceDesc::<H, W> {
    //             functional_units,
    //             switchboxes,
    //         },
    //         live_outs,
    //     ))
    // }
}

fn write_xdot<const H: usize, const W: usize, T: Write>(slice: &SliceDesc<H, W>, writer: &mut T)
where
    [(); H]:,
    [[Switchbox<W>; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
    [[FunctionalUnit; W]; H]: Serialize + for<'d2> Deserialize<'d2>,
{
    // writeln!(writer,"XDOT:");
    writeln!(writer, "digraph {{");
    writeln!(writer, "node [shape=square]");
    for (row_num, row) in slice.functional_units.iter().enumerate() {
        for (col_num, fu) in row.iter().enumerate() {
            //if fu.op_type != FuOp::Default {
            writeln!(
                writer,
                "n_{}_{} [label=\"{:?}{:?}\"]",
                row_num, col_num, fu.op_type, col_num
            );
            //}
        }
    }

    for (row_num, row) in slice.switchboxes.iter().enumerate() {
        for (col_num, sb) in row.iter().enumerate() {
            if sb.output_wires.0 != 100 {
                writeln!(
                    writer,
                    "n_{}_{} -> n_{}_{}",
                    row_num,
                    sb.output_wires.0,
                    row_num + 1,
                    col_num
                );
            }

            if sb.output_wires.1 != 100 {
                if row_num < H - 1 {
                    if slice.functional_units[row_num + 1][col_num].op_type != FuOp::PassA {
                        writeln!(
                            writer,
                            "n_{}_{} -> n_{}_{}",
                            row_num,
                            sb.output_wires.1,
                            row_num + 1,
                            col_num
                        );
                    }
                } else {
                    writeln!(
                        writer,
                        "n_{}_{} -> n_{}_{}",
                        row_num,
                        sb.output_wires.1,
                        row_num + 1,
                        col_num
                    );
                }
            }
        }
    }

    writeln!(writer, "edge [weight=1000 style=invis]");
    for (row_num, row) in slice.functional_units.iter().enumerate() {
        for col_num in 0..row.len() {
            write!(writer, "n_{}_{}", col_num, row_num);
            if col_num + 1 < row.len() {
                write!(writer, " -> ");
            }
        }
        writeln!(writer, "");
    }
    for (row_num, row) in slice.functional_units.iter().enumerate() {
        write!(writer, "rank=same {{");
        for col_num in 0..row.len() {
            write!(writer, "n_{}_{}", row_num, col_num);
            if col_num + 1 < row.len() {
                write!(writer, " -> ");
            }
        }
        writeln!(writer, "}}");
    }
    writeln!(writer, "}}");
}
